/*
 * Copyright (c) 2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jul  6 2016
 */
#include "acdynamixel.h"

#include <sys/time.h>
#include <err.h>
#include <unistd.h>

#include "dynamixel_c_types.h"
#include "codels.h"


/* --- Task comm -------------------------------------------------------- */


/** Codel dx_comm_start of task comm.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_send.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_start(dynamixel_ids *ids, const dynamixel_motors *motors,
              genom_context self)
{
  or_joint_state *mdata;
  int i;

  ids->chan = malloc(sizeof(*ids->chan));
  if (!ids->chan) return dx_e_sys_error(NULL, self);

  ids->chan->fd = -1;
  ids->chan->r = ids->chan->w = 0;
  ids->input.mode = dynamixel_cmd_position;
  ids->input.position._length = 0;
  ids->input.velocity._length = 0;
  ids->input.effort._length = 0;

  mdata = motors->data(self);
  if (mdata) {
    mdata->ts.sec = mdata->ts.nsec = 0;
    mdata->name._length = 0;
    mdata->flag._length = mdata->flag._maximum;
    mdata->position._length = 0;
    mdata->velocity._length = 0;
    mdata->effort._length = 0;
    for(i = 0; i < mdata->flag._maximum; i++)
      mdata->flag._buffer[i].active = false;
  }

  return dynamixel_send;
}


/** Codel dx_comm_send of task comm.
 *
 * Triggered by dynamixel_send.
 * Yields to dynamixel_pause_send, dynamixel_poll.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_send(dynamixel_channel_s **chan, dynamixel_ids_input_s *input,
             genom_context self)
{
  int i;

  if ((*chan)->fd < 0) return dynamixel_pause_send;

  /* send position target */
  if (input->position._length > 0) {
    uint8_t p[2 * input->position._length];
    int16_t v;

    if (input->mode != dynamixel_cmd_position) {
      dx_send_msg(*chan, 0xfe, DX_WRITE,
                  4, DX_CW_ANGLE_LIMIT, 0xff, 0xf, 0xff, 0xf);
      dx_send_msg(*chan, 0xfe, DX_WRITE, 2, DX_MOVING_SPEED, 0, 0);
      dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_TORQUE_CONTROL_MODE, 0);
      input->mode = dynamixel_cmd_position;
    }

    /* send position target */
    for(i = 0; i < input->position._length; i++) {
      v = (input->position._buffer[i] + M_PI) / DX_POSITION_RESOLUTION;
      p[2*i] = v & 0xff;
      p[2*i+1] = v >> 8;
    }

    dx_send_msg(*chan, 0xfe, DX_SYNC_WRITE, input->position._length,
                2, DX_GOAL_POSITION, p);
  }


  /* send velocity target */
  if (input->velocity._length > 0) {
    uint8_t p[2 * input->velocity._length];
    int16_t v;

    if (input->mode != dynamixel_cmd_velocity) {
      dx_send_msg(*chan, 0xfe, DX_WRITE, 4, DX_CW_ANGLE_LIMIT, 0, 0, 0, 0);
      dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_TORQUE_CONTROL_MODE, 0);
      input->mode = dynamixel_cmd_velocity;
    }

    for(i = 0; i < input->velocity._length; i++) {
      v = input->velocity._buffer[i] / DX_VELOCITY_RESOLUTION;

      if (v > 1023) v = 1023;
      if (v < -1023) v = -1023;
      if (v < 0) v = 1023 - v;

      p[2*i] = v & 0xff;
      p[2*i+1] = v >> 8;
    }

    dx_send_msg(*chan, 0xfe, DX_SYNC_WRITE, input->velocity._length,
                  2, DX_MOVING_SPEED, p);
  }


  /* send effort target */
  if (input->effort._length > 0) {
    uint8_t p[2 * input->effort._length];
    int16_t v;

    if (input->mode != dynamixel_cmd_effort) {
      dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_TORQUE_CONTROL_MODE, 1);
      input->mode = dynamixel_cmd_effort;
    }

    for(i = 0; i < input->effort._length; i++) {
      v = input->effort._buffer[i] / DX_TORQUE_RESOLUTION;

      if (v > 1023) v = 1023;
      if (v < -1023) v = -1023;
      if (v < 0) v = 1023 - v;

      p[2*i] = v & 0xff;
      p[2*i+1] = v >> 8;
    }

    dx_send_msg(*chan, 0xfe, DX_SYNC_WRITE, input->effort._length,
                2, DX_GOAL_TORQUE, p);
  }


  /* request data from all motors */
  dx_send_msg(*chan, 0xfe, DX_BULK_READ,
              or_joint_max_joints, 6, DX_PRESENT_POSITION);

  return dynamixel_poll;
}


/** Codel dx_comm_poll of task comm.
 *
 * Triggered by dynamixel_poll.
 * Yields to dynamixel_recv, dynamixel_send.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_poll(const dynamixel_channel_s *chan, genom_context self)
{
  int s;

  s = dx_wait_msg(chan, dynamixel_control_period_ms);
  switch(s) {
    case -2:
      warnx("disconnected from %s", chan->path);
      close(chan->fd);
      ((dynamixel_channel_s *)chan)->fd = -1;
      return dynamixel_send;
    case -1: case 0:
      return dynamixel_send;
  }
  return dynamixel_recv;
}


/** Codel dx_comm_recv of task comm.
 *
 * Triggered by dynamixel_recv.
 * Yields to dynamixel_pause_send.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_recv(dynamixel_channel_s **chan,
             const dynamixel_motors *motors, genom_context self)
{
  or_joint_state *mdata = motors->data(self);
  struct timeval tv;
  int16_t v;
  int id;

  gettimeofday(&tv, NULL);
  for(id = 0; id < mdata->flag._length; id++)
    mdata->flag._buffer[id-1].active = false;

  while(dx_recv_msg(*chan, false) == 1) {
    if ((*chan)->p != 6) continue;
    id = (*chan)->id;
    if (id >= or_joint_max_joints) continue;

    if (mdata) {
      mdata->ts.sec = tv.tv_sec;
      mdata->ts.nsec = 1000 * tv.tv_usec;

      mdata->flag._buffer[id-1].active = true;

      if (mdata->position._length < id) mdata->position._length = id;
      if (mdata->velocity._length < id) mdata->velocity._length = id;
      if (mdata->effort._length < id) mdata->effort._length = id;

      v = (*chan)->param[0] + ((*chan)->param[1] << 8);
      mdata->position._buffer[id-1] = v * DX_POSITION_RESOLUTION - M_PI;

      v = ((*chan)->param[2] + ((*chan)->param[3] << 8)) & 0x3ff;
      if ((*chan)->param[3] & 4) v = -v;
      mdata->velocity._buffer[id-1] = v * DX_VELOCITY_RESOLUTION;

      v = ((*chan)->param[4] + ((*chan)->param[5] << 8)) & 0x3ff;
      if ((*chan)->param[5] & 4) v = -v;
      mdata->effort._buffer[id-1] = v * DX_TORQUE_RESOLUTION;
    }
  }

  motors->write(self);
  return dynamixel_pause_send;
}


/** Codel dx_comm_stop of task comm.
 *
 * Triggered by dynamixel_stop.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_stop(dynamixel_channel_s **chan, genom_context self)
{
  /* skeleton sample: insert your code */
  /* skeleton sample */ return dynamixel_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel dx_connect_start of activity connect.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_send.
 * Throws dynamixel_e_sys, dynamixel_e_baddev.
 */
genom_event
dx_connect_start(const char serial[64], uint32_t baud,
                 dynamixel_channel_s **chan, genom_context self)
{
  int s, c;

  if ((*chan)->fd >= 0) {
    close((*chan)->fd);
    warnx("disconnected from %s", (*chan)->path);
  }

  /* open tty */
  (*chan)->fd = dx_open_tty(serial, baud);
  if ((*chan)->fd < 0) return dx_e_sys_error(serial, self);
  snprintf((*chan)->path, sizeof((*chan)->path), "%s", serial);
  (*chan)->state = 0;

  /* check endpoint */
  while (dx_recv_msg(*chan, true) == 1); /* flush buffer */

  c = 0;
  do {
    do {
      if (dx_send_msg(*chan, 0xfe, DX_PING)) { /* broadcast ping */
        close((*chan)->fd);
        (*chan)->fd = -1;
        return dx_e_sys_error(serial, self);
      }

      s = dx_wait_msg(*chan, 100/*ms*/);
      if (s < 0 && errno != EINTR) {
        close((*chan)->fd);
        (*chan)->fd = -1;
        return dx_e_sys_error(NULL, self);
      }
    } while(s <= 0);

    s = dx_recv_msg(*chan, true);
  } while(s != 1 && c++ < 3);
  while (dx_recv_msg(*chan, true) == 1); /* flush buffer */

  if (c > 3) {
    close((*chan)->fd);
    (*chan)->fd = -1;
    errno = ETIMEDOUT;
    return dx_e_sys_error(NULL, self);
  }

  return dynamixel_send;
}

/** Codel dx_connect_configure of activity connect.
 *
 * Triggered by dynamixel_send.
 * Yields to dynamixel_poll.
 * Throws dynamixel_e_sys, dynamixel_e_baddev.
 */
genom_event
dx_connect_configure(dynamixel_channel_s **chan, genom_context self)
{
  /* disable status for writes */
  dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_STATUS_RETURN_LEVEL, 1);

  /* disable delay */
  dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_RETURN_DELAY_TIME, 0);

  /* set position control mode */
  dx_send_msg(*chan, 0xfe, DX_WRITE,
              4, DX_CW_ANGLE_LIMIT, 0xff, 0xf, 0xff, 0xf);
  dx_send_msg(*chan, 0xfe, DX_WRITE, 2, DX_MOVING_SPEED, 0, 0);
  dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_TORQUE_CONTROL_MODE, 0);

  /* get model id */
  dx_send_msg(*chan, 0xfe, DX_BULK_READ,
              or_joint_max_joints, 3, DX_MODEL_NUMBER);

  return dynamixel_poll;
}

/** Codel dx_comm_poll of activity connect.
 *
 * Triggered by dynamixel_poll.
 * Yields to dynamixel_recv, dynamixel_send.
 * Throws dynamixel_e_sys, dynamixel_e_baddev.
 */
/* already defined in task comm */


/** Codel dx_connect_recv of activity connect.
 *
 * Triggered by dynamixel_recv.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys, dynamixel_e_baddev.
 */
genom_event
dx_connect_recv(dynamixel_channel_s **chan,
                const dynamixel_motors *motors, genom_context self)
{
  static const struct { int id; const char *name; } models[2] = {
    { .id = 0x1d, .name = "MX-28" },
    { .id = 0x136, .name = "MX-64" }
  };

  or_joint_state *mdata = motors->data(self);
  int model, id;
  int i;

  while(dx_recv_msg(*chan, true) == 1) {
    if ((*chan)->p != 3) continue;
    id = (*chan)->id;
    if (id >= or_joint_max_joints) continue;

    if (mdata->name._length < id) mdata->name._length = id;
    snprintf(mdata->name._buffer[id - 1],
             sizeof(mdata->name._buffer[id - 1]), "unknown (version %d)",
             (*chan)->param[2]);

    model = (*chan)->param[0] + ((*chan)->param[1] << 8);
    for(i = 0; i < sizeof(models)/sizeof(models[0]); i++) {
      if (models[i].id == model) {
        snprintf(mdata->name._buffer[id - 1],
                 sizeof(mdata->name._buffer[id - 1]), "%s (version %d)",
                 models[i].name, (*chan)->param[2]);
        break;
      }
    }

    warnx("detected motor %d: %s", id, mdata->name._buffer[id - 1]);
  }

  motors->write(self);

  warnx("connected to %s", (*chan)->path);
  return dynamixel_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel dx_disconnect_start of activity disconnect.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_disconnect_start(dynamixel_channel_s **chan, genom_context self)
{
  if ((*chan)->fd >= 0) {
    close((*chan)->fd);
    (*chan)->fd = -1;
    warnx("disconnected from %s", (*chan)->path);
  }

  return dynamixel_ether;
}
