/*
 * Copyright (c) 2015-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Wed Jul  6 2016
 */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include "codels.h"


/* --- dx_open_tty --------------------------------------------------------- */

/* Open a serial port, configure to the given baud rate */
int
dx_open_tty(const char *device, uint32_t speed)
{
  struct termios t;
  speed_t baud;
  int fd;

  /* select baud rate */
#ifndef B57600
# define B57600 57600U
#endif
#ifndef B115200
# define B115200 115200U
#endif
#ifndef B500000
# define B500000 500000U
#endif
#ifndef B1000000
# define B1000000 1000000U
#endif
#ifndef B2000000
# define B2000000 2000000U
#endif
#ifndef B2000000
# define B3000000 3000000U
#endif
  switch(speed) {
    case 57600:		baud = B57600; break;
    case 115200:	baud = B115200; break;
    case 500000:	baud = B500000; break;
    case 1000000:	baud = B1000000; break;
    case 2000000:	baud = B2000000; break;
    case 3000000:	baud = B3000000; break;

    default: errno = EINVAL; return -1;
  }

  /* open non-blocking */
  fd = open(device, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd < 0) return fd;
  if (!isatty(fd)) {
    errno = ENOTTY;
    return -1;
  }

  /* configure line discipline */
  if (tcgetattr(fd, &t)) return -1;

  t.c_iflag = IGNBRK;
  t.c_oflag = 0;
  t.c_lflag = 0;
  t.c_cflag = CS8 | CREAD | CLOCAL;
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 0;

  if (cfsetospeed(&t, baud)) return -1;
  if (cfsetispeed(&t, baud)) return -1;

  if (tcsetattr(fd, TCSANOW, &t)) return -1;

  /* discard any pending data */
  tcflush(fd, TCIOFLUSH);

  return fd;
}


/* --- dx_wait_msg --------------------------------------------------------- */

int
dx_wait_msg(const struct dynamixel_channel_s *channel, int timeout)
{
  struct pollfd pfd;
  int s;

  pfd.fd = channel->fd;
  pfd.events = POLLIN;

  s = poll(&pfd, pfd.fd < 0 ? 0 : 1, timeout);
  if (s < 0 && errno == EINTR) return 0;
  if (s > 0 && pfd.revents & POLLHUP) return -2;

  return s;
}


/* --- dx_recv_msg --------------------------------------------------------- */

/* returns: 0: timeout/incomplete, -1: error, 1: complete msg */

int
dx_recv_msg(struct dynamixel_channel_s *chan, bool block)
{
  struct iovec iov[2];
  ssize_t s;
  uint8_t c;

  if (chan->fd < 0) return -1;

  do {
    /* feed the ring  buffer */
    iov[0].iov_base = chan->buf + chan->w;
    iov[1].iov_base = chan->buf;

    if (chan->r > chan->w) {
      iov[0].iov_len = chan->r - chan->w - 1;
      iov[1].iov_len = 0;
    } else if (chan->r > 0) {
      iov[0].iov_len = sizeof(chan->buf) - chan->w;
      iov[1].iov_len = chan->r - 1;
    } else {
      iov[0].iov_len = sizeof(chan->buf) - chan->w - 1;
      iov[1].iov_len = 0;
    }

    if (iov[0].iov_len || iov[1].iov_len) {
      s = readv(chan->fd, iov, 2);

      if (s < 0) return -1;
      else if (s == 0 && block) {
        struct pollfd fd = { .fd = chan->fd, .events = POLLIN };

        s = poll(&fd, 1, 100/*ms*/);
        if (s >= 0 && fd.revents & POLLHUP) return -2;
      } else
        chan->w = (chan->w + s) % sizeof(chan->buf);

      if (s == 0 && chan->r == chan->w) return 0;
    }

    while(chan->r != chan->w) {
      c = chan->buf[chan->r];
      chan->r = (chan->r + 1) % sizeof(chan->buf);

      switch(chan->state) {
        case 0: case 1:
          if (c == 0xff) chan->state++; else chan->state = 0;
          break;

        case 2:
          if (c == 0xff) { chan->state = 1; break; }
          chan->id = c; chan->sum = c;
          chan->state++; break;

        case 3:
          if (c == 0xff) { chan->state = 1; break; }
          chan->len = (c > 1) ? c - 2 : 0; chan->sum += c;
          chan->p = 0;
          chan->state++; break;
        case 4:
          if (c == 0xff) { chan->state = 1; break; }
          chan->error = c; chan->sum += c;
          chan->state++; break;

        default:
          if (chan->len) {
            if (chan->p > 0 && chan->param[chan->p-1] == 0xff && c == 0xff) {
              chan->state = 2; break;
            }
            chan->param[chan->p++] = c; chan->sum += c;
            chan->len--;
          } else {
            chan->state = 0;
            if (c == (uint8_t)~chan->sum) return 1;
          }
          break;
      }
    }
  } while(1);

  return 0;
}


/* --- dx_send_msg --------------------------------------------------------- */

int
dx_send_msg(const struct dynamixel_channel_s *chan,
            uint8_t id, enum dx_cmd cmd, ...)
{
  va_list ap;
  ssize_t s;
  char buf[64], *w, *r;
  uint8_t cksum;

  if (chan->fd < 0) return -1;

  /* encode message */
  va_start(ap, cmd);
  w = buf;
  *w++ = 0xff;
  *w++ = 0xff;
  *w++ = id; cksum = id;
  switch(cmd) {
    case DX_PING:
      *w++ = 2; cksum += 2;
      *w++ = cmd; cksum += cmd;
      break;

    case DX_READ: {
      uint8_t n = va_arg(ap, int);
      uint8_t addr = va_arg(ap, int);
      *w++ = 4; cksum += 4;
      *w++ = cmd; cksum += cmd;
      *w++ = addr; cksum += addr;
      *w++ = n; cksum += n;
      break;
    }
    case DX_BULK_READ: {
      int ids = va_arg(ap, int);
      uint8_t n = va_arg(ap, int);
      uint8_t addr = va_arg(ap, int);
      int i;

      *w++ = 3*ids + 3; cksum += 3*ids + 3;
      *w++ = cmd; cksum += cmd;
      *w++ = 0;
      for(i = 1; i <= ids; i++) {
        *w++ = n; cksum += n;
        *w++ = i; cksum += i;
        *w++ = addr; cksum += addr;
      }
      break;
    }

    case DX_WRITE: {
      uint8_t d, i, n = va_arg(ap, int);
      *w++ = 3 + n; cksum += 3 + n;
      *w++ = cmd; cksum += cmd;
      for(i = 0; i < n + 1; i++) {
        d = va_arg(ap, int);
        *w++ = d; cksum += d;
      }
      break;
    }
    case DX_SYNC_WRITE: {
      int ids = va_arg(ap, int);
      uint8_t n = va_arg(ap, int);
      uint8_t addr = va_arg(ap, int);
      uint8_t *d = va_arg(ap, uint8_t *);
      int i, j;

      *w++ = (n+1)*ids + 4; cksum += (n+1)*ids + 4;
      *w++ = cmd; cksum += cmd;
      *w++ = addr; cksum += addr;
      *w++ = n; cksum += n;
      for(i = 1; i <= ids; i++) {
        *w++ = i; cksum += i;
        for(j = 0; j < n; j++) {
          *w++ = *d; cksum += *d; d++;
        }
      }
      break;
    }
  }
  *w++ = ~cksum;
  va_end(ap);

  /* send */
  r = buf;
  while (w > r) {
    do {
      s = write(chan->fd, r, w - r);
    } while (s < 0 && errno == EINTR);
    if (s < 0) return -1;

    r += s;
  }

  /* add a small delay after writes, to not confuse the controller */
  {
    struct timespec req = { 0, 300000 };
    nanosleep(&req, NULL);
  }

  return 0;
}
